
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "nrf_log.h"
#include "nrf_soc.h"
#include "nrf_sdm.h"
#include "nrf_sdh.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"
#include "nrf_strerror.h"

#include "app_error.h"
#include "ble.h"
#include "ble_srv_common.h"

#include "GhsControlStructs.h"

const char * getErrorString(unsigned long err_code);


/* Usage:
 *  char bigString[1000];
 *  char *p = bigString;
 *  bigString[0] = '\0';
 *  p = xstrcat(p,"John, ");
 *  p = xstrcat(p,"Paul, ");
 *  p = xstrcat(p,"George, ");
 *  p = xstrcat(p,"Joel ");
 */
char* xstrcat(char* dest, const char* src)
{
    while (*dest) dest++;       // Increment the contents until it is 0 (string end)
    while (*dest++ = *src++);   // contents of src extracted, src incremented, place into contents of dest, dest incremented
    return --dest;
}

char* byteToHex(unsigned char* bytes, char* buffer, char* separator, unsigned long count)
{
    char* cptr = buffer;
    unsigned long i;
    char buf[4];
    *cptr = 0;
    for (i = 0; i < count - 1; i++)
    {
        sprintf(buf, "%02X", bytes[i]);
        cptr = xstrcat(cptr, buf);
        cptr = xstrcat(cptr, separator);
    }
    sprintf(buf, "%02X", bytes[i]);
    cptr = xstrcat(cptr, buf);
    return buffer;
}

int twoByteEncode(unsigned char* msmtBuf, int index, unsigned short value)
{
    msmtBuf[index++] = (unsigned char)(value & 0xFF);
    msmtBuf[index++] = (unsigned char)((value >> 8) & 0xFF);
    return index;
}

int fourByteEncode(unsigned char* msmtBuf, int index, unsigned long value)
{
    int m;
    for (m = 0; m < 4; m++)
    {
        msmtBuf[index++] = (unsigned char)(value & 0xFF);
        value = (value >> 8);
    }
    return index;
}

bool hexToLittleEndianByte(char* hexString, unsigned char* byteArray)
{

    if ((hexString == NULL) || (byteArray == NULL))
    {
        NRF_LOG_DEBUG("One or both of the input parameters are NULL");
        return false;
    }
    size_t length = strlen(hexString);
    if ((length & 1) == 1)
    {
        NRF_LOG_DEBUG("Hex String length is not even\n");
        return false;
    }
    for (int i = 0; i < length; i++)
    {
        if ((hexString[i] >= 0x30) && hexString[i] <= 0x39)
        {
            continue;
        }
        if ((hexString[i] >= 'A') && hexString[i] <= 'F')
        {
            continue;
        }
        if ((hexString[i] >= 'a') && hexString[i] <= 'f')
        {
            continue;
        }
        NRF_LOG_DEBUG("Not a legal HEX string");
        return false;
    }
    const char* pos = hexString;
    length = (length >> 1);

    for (size_t count = 1; count <= length; count++)
    {
        if (sscanf(pos, "%2hhx", &byteArray[length - count]) < 0)
        {
            NRF_LOG_DEBUG("Error converting the HEX string element at %d to a byte", (pos - hexString));
            return false;
        }
        pos += 2;
    }

    NRF_LOG_DEBUG("0x");
    for (size_t count = 0; count < length; count++)
    {
        NRF_LOG_DEBUG("%02x ", byteArray[count]);
    }
    NRF_LOG_DEBUG("");

    return true;
}


// Generic method to create a primary service
ret_code_t createPrimaryService(unsigned short* serviceHandle, unsigned short serviceUuid)
{
    ret_code_t    error_code;
    ble_uuid_t  ble_uuid;

    BLE_UUID_BLE_ASSIGN(ble_uuid, serviceUuid);

    error_code = sd_ble_gatts_service_add(
        BLE_GATTS_SRVC_TYPE_PRIMARY,
        &ble_uuid,
        serviceHandle);

    if (error_code != NRF_SUCCESS)
    {
        NRF_LOG_ERROR("Failed to initialize service. Error: %s (0x%04X)", getErrorString(error_code), error_code);
        return error_code;
    }
    NRF_LOG_DEBUG("Service initiated");
    return NRF_SUCCESS;
}

// Generic method to create a characteristic and add it to a service
ret_code_t createStandardCharacteristic(
    uint16_t serviceHandle,
    ble_gatts_char_handles_t* charHandles,
    uint16_t uuid,
    bool hasCCCD,
    bool indicate,
    uint16_t valueLength,
    uint8_t *value,
    bool trapRead,
    ble_gap_conn_sec_mode_t cccdWritePermission,
    ble_gap_conn_sec_mode_t attrReadPermission,
    ble_gap_conn_sec_mode_t attrWritePermission,
    bool isStatic)
{
    ret_code_t           error_code;
    ble_gatts_char_md_t  char_md;
    ble_gatts_attr_md_t  cccd_md;
    ble_gatts_attr_md_t* cccd_md_ptr = NULL;
    ble_gatts_attr_t     attr_char_value;
    ble_uuid_t           ble_uuid;
    ble_gatts_attr_md_t  attr_md;
    uint8_t              dummy = 1;

    if (hasCCCD)
    {
        memset(&cccd_md, 0, sizeof(cccd_md));

        cccd_md.rd_auth = 0;
        BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);
        cccd_md.write_perm.lv = cccdWritePermission.lv;
        cccd_md.write_perm.sm = cccdWritePermission.sm;
        cccd_md.vloc = BLE_GATTS_VLOC_STACK;
        cccd_md_ptr = &cccd_md;
    }

    memset(&char_md, 0, sizeof(char_md));

    if (hasCCCD)
    {
        if (indicate)
        {
            char_md.char_props.indicate = 1;
        }
        else
        {
            char_md.char_props.notify = 1;
        }
    }
    char_md.char_props.read = ((attrReadPermission.sm == 0) ? 0 : 1);
    char_md.char_props.write = ((attrWritePermission.sm == 0) ? 0 : 1);
    char_md.p_char_user_desc = NULL;
    char_md.p_char_pf = NULL;
    char_md.p_user_desc_md = NULL;
    char_md.p_cccd_md = cccd_md_ptr;
    char_md.p_sccd_md = NULL;

    BLE_UUID_BLE_ASSIGN(ble_uuid, uuid);

    memset(&attr_md, 0, sizeof(attr_md));

    attr_md.read_perm.sm = attrReadPermission.sm;
    attr_md.read_perm.lv = attrReadPermission.lv;
    attr_md.write_perm.sm = attrWritePermission.sm;
    attr_md.write_perm.lv = attrWritePermission.lv;

    attr_md.vloc = BLE_GATTS_VLOC_STACK;
    attr_md.rd_auth = trapRead ? 1 : 0;
    attr_md.wr_auth = 0;
    attr_md.vlen = 1;

    memset(&attr_char_value, 0, sizeof(attr_char_value));

    attr_char_value.p_uuid = &ble_uuid;
    attr_char_value.p_attr_md = &attr_md;
    attr_char_value.init_len = (valueLength == 0) ? 1 : valueLength; //attr_char_value_init_len;
    attr_char_value.init_offs = 0;
    attr_char_value.max_len = isStatic? (valueLength + OPCODE_LENGTH + HANDLE_LENGTH) : (NRF_SDH_BLE_GAP_DATA_LENGTH - 4 - OPCODE_LENGTH - HANDLE_LENGTH);
    if ((value == NULL) || (valueLength == 0))
    {
        attr_char_value.p_value = &dummy;    // Fill with any garbage
    }
    else
    {
        NRF_LOG_DEBUG("char val length %u", valueLength);
        attr_char_value.p_value = value;
    }

    error_code = sd_ble_gatts_characteristic_add(serviceHandle,
        &char_md,
        &attr_char_value,
        charHandles);

    if (error_code != NRF_SUCCESS)
    {
        NRF_LOG_ERROR("Failed to initialize characteristics. Error: %s (0x%04X)", getErrorString(error_code), error_code);
        return error_code;
    }
    NRF_LOG_DEBUG("Characteristic 0x%04X initiated", uuid);

    return NRF_SUCCESS;
}

static ble_gatts_char_handles_t m_systemId_handle;
static ble_gatts_char_handles_t m_manufacturer_handle;
static ble_gatts_char_handles_t m_model_handle;
static ble_gatts_char_handles_t m_serial_handle;
static ble_gatts_char_handles_t m_firmware_handle;
static ble_gatts_char_handles_t m_software_handle;
static ble_gatts_char_handles_t m_hardware_handle;
static ble_gatts_char_handles_t m_regcert_handle;
ret_code_t createDeviceInformationService(unsigned short int *serviceHandle, s_DisCharValues *disCharValues)
{
    ret_code_t error_code = createPrimaryService(serviceHandle, BTLE_DEVICE_INFORMATION_SERVICE);
    if (error_code != NRF_SUCCESS)
    {
        NRF_LOG_ERROR("Unable to create Device Information Service. Error: %s (0x%04X)", getErrorString(error_code), error_code);
        return error_code;
    }


        error_code = createStandardCharacteristic(*serviceHandle,
                &m_systemId_handle,
                BTLE_DIS_SYSTEM_ID_CHAR,
                false,
                false,
                8,
                disCharValues->systemId,
                disCharValues->trapRead,
                (ble_gap_conn_sec_mode_t) {    0, 0},
                (ble_gap_conn_sec_mode_t) {    1, 1},
                (ble_gap_conn_sec_mode_t) {    0, 0},
                true);

        if (error_code != NRF_SUCCESS)
        {
                NRF_LOG_ERROR("Unable to create System Id characteristic. Error: %s (0x%04X)", getErrorString(error_code), error_code);
        }

    if (disCharValues->manufacturerName != NULL)
    {
        error_code = createStandardCharacteristic(*serviceHandle,
            &m_manufacturer_handle,
            BTLE_DIS_MANUFACTURER_DEVICE_NAME_CHAR,
            false,
            false,
            (uint16_t)strlen(disCharValues->manufacturerName),
            (uint8_t *)disCharValues->manufacturerName,
            disCharValues->trapRead,
            (ble_gap_conn_sec_mode_t) {    0, 0},
            (ble_gap_conn_sec_mode_t) {    1, 1},
            (ble_gap_conn_sec_mode_t) {    0, 0},
                        true);

        if (error_code != NRF_SUCCESS)
        {
            NRF_LOG_ERROR("Unable to create Manufacturer Name characteristic. Error: %s (0x%04X)", getErrorString(error_code), error_code);
        }
    }
    if (disCharValues->modelNumber != NULL)
    {
        error_code = createStandardCharacteristic(*serviceHandle,
            &m_model_handle,
            BTLE_DIS_MODEL_NUMBER_CHAR,
            false,
            false,
            (uint16_t)strlen(disCharValues->modelNumber),
            (uint8_t *)disCharValues->modelNumber,
            disCharValues->trapRead,
            (ble_gap_conn_sec_mode_t) {    0, 0},
            (ble_gap_conn_sec_mode_t) {    1, 1},
            (ble_gap_conn_sec_mode_t) {    0, 0},
                        true);

        if (error_code != NRF_SUCCESS)
        {
            NRF_LOG_ERROR("Unable to create Model Number characteristic. Error: %s (0x%04X)", getErrorString(error_code), error_code);
        }
    }
    if (disCharValues->serialNumber != NULL)
    {
        error_code = createStandardCharacteristic(*serviceHandle,
            &m_serial_handle,
            BTLE_DIS_SERIAL_NUMBER_CHAR,
            false,
            false,
            (uint16_t)strlen(disCharValues->serialNumber),
            (uint8_t *)disCharValues->serialNumber,
            disCharValues->trapRead,
            (ble_gap_conn_sec_mode_t) {    0, 0},
            (ble_gap_conn_sec_mode_t) {    1, 1},
            (ble_gap_conn_sec_mode_t) {    0, 0},
                        true);

        if (error_code != NRF_SUCCESS)
        {
            NRF_LOG_ERROR("Unable to create Serial Number characteristic. Error: %s (0x%04X)", getErrorString(error_code), error_code);
        }
    }
    if (disCharValues->firmwareRevision != NULL)
    {
        error_code = createStandardCharacteristic(*serviceHandle,
            &m_firmware_handle,
            BTLE_DIS_FIRMWARE_REVISION_CHAR,
            false,
            false,
            (uint16_t)strlen(disCharValues->firmwareRevision),
            (uint8_t *)disCharValues->firmwareRevision,
            disCharValues->trapRead,
            (ble_gap_conn_sec_mode_t) {    0, 0},
            (ble_gap_conn_sec_mode_t) {    1, 1},
            (ble_gap_conn_sec_mode_t) {    0, 0},
                        true);

        if (error_code != NRF_SUCCESS)
        {
            NRF_LOG_ERROR("Unable to create Firmware Revision characteristic. Error: %s (0x%04X)", getErrorString(error_code), error_code);
        }
    }
    if (disCharValues->hardwareRevision != NULL)
    {
        error_code = createStandardCharacteristic(*serviceHandle,
            &m_hardware_handle,
            BTLE_DIS_HARDWARE_REVISION_CHAR,
            false,
            false,
            (uint16_t)strlen(disCharValues->hardwareRevision),
            (uint8_t *)disCharValues->hardwareRevision,
            disCharValues->trapRead,
            (ble_gap_conn_sec_mode_t) {    0, 0},
            (ble_gap_conn_sec_mode_t) {    1, 1},
            (ble_gap_conn_sec_mode_t) {    0, 0},
                        true);

        if (error_code != NRF_SUCCESS)
        {
            NRF_LOG_ERROR("Unable to create Hardware Revision characteristic. Error: %s (0x%04X)", getErrorString(error_code), error_code);
        }
    }
    if (disCharValues->softwareRevision != NULL)
    {
        error_code = createStandardCharacteristic(*serviceHandle,
            &m_software_handle,
            BTLE_DIS_SOFTWARE_REVISION_CHAR,
            false,
            false,
            (uint16_t)strlen(disCharValues->softwareRevision),
            (uint8_t *)disCharValues->softwareRevision,
            disCharValues->trapRead,
            (ble_gap_conn_sec_mode_t) {    0, 0},
            (ble_gap_conn_sec_mode_t) {    1, 1},
            (ble_gap_conn_sec_mode_t) {    0, 0},
                        true);

        if (error_code != NRF_SUCCESS)
        {
            NRF_LOG_ERROR("Unable to create Software Revision characteristic. Error: %s (0x%04X)", getErrorString(error_code), error_code);
        }
    }
    if (disCharValues->regCertDataList != NULL)
    {
        NRF_LOG_DEBUG("Reg cert Max size %u\n", MAX_CHAR_LEN);
        error_code = createStandardCharacteristic(*serviceHandle,
            &m_regcert_handle,
            BTLE_DIS_REG_CERT_DATA_LIST_CHAR,
            false,
            false,
            disCharValues->regCertDataListLength,
            disCharValues->regCertDataList,
            disCharValues->trapRead,
            (ble_gap_conn_sec_mode_t) {    0, 0},
            (ble_gap_conn_sec_mode_t) {    1, 1},
            (ble_gap_conn_sec_mode_t) {    0, 0},
                        true);

        if (error_code != NRF_SUCCESS)
        {
            NRF_LOG_ERROR("Unable to create RegCertDataList characteristic. Error: %s (0x%04X)", getErrorString(error_code), error_code);
        }
    }
    return NRF_SUCCESS;
}

static const char nameKey[7] = {'C', 'Y', 'C', '_', 'P', 'W', 'R'}; // Needed because we use same board for multiple projects
void saveKeysToFlash(ble_gap_sec_keyset_t* keys,
    unsigned char **saveDataBuffer, unsigned short int *saveDataLength,
    bool* cccdSet, unsigned short* noOfCccds)
{
    int i;
//        char buffer[1024];
    unsigned char *keysDataBuffer = NULL;
    ret_code_t err_code;

    uint32_t pg_size = NRF_FICR->CODEPAGESIZE;
    uint32_t pg_num  = NRF_FICR->CODESIZE - (NRF_FICR->CODESIZE * pg_size - 0xE0000) / pg_size;
    
    // Okay, what are we doing here. On the dongle, the bootloader occupies flash above 0xE0000. So I have to find how big flash is,
    // subtract 0xE0000 from it, find the number of pages there, and then subtract that number of pages from the total. THAT gives the
    // page where the boot record starts (0xE0000). Now I have to drop a few pages to where I can start. I drop two pages here which is
    // more than enough room for this app. I still don't know how to dynamically find the page at the top of MY app which gives the
    // area of free flash between my app and the bootloader at the top. The DK, by the way, has no bootloader so I don't need all this.

    NRF_LOG_INFO("Total pages is %u, pg_num is %u pg_size is %u", NRF_FICR->CODESIZE, pg_num, NRF_FICR->CODEPAGESIZE);
    pg_num = pg_num - 2;
    uint32_t *addr;
    NRF_LOG_DEBUG("Saving data to flash.");

    int size = sizeof(nameKey) +
               sizeof(ble_gap_enc_key_t) +      // should be a fixed length even if no pairing was done
               sizeof(ble_gap_id_key_t) +       // should be a fixed length even if no pairing was done so no IRK
               sizeof(ble_gap_enc_key_t) +
               sizeof(ble_gap_id_key_t) +
               sizeof(unsigned short) +         // Number of CCCDs
               sizeof(bool) * (*noOfCccds) +    // CCCD set
               sizeof(unsigned short) +         // Savedata length
               ((*saveDataBuffer != NULL) ? *saveDataLength * sizeof(unsigned char) : 0);
    // The writes are done in 4-byte hunks so we have to even out the length
    size = 4 + ((size >> 2) << 2);  // Here we round up to the nearest 4-byte size so it is evenly divisible by 4
    NRF_LOG_DEBUG("Size of data to write %u.", size);
    keysDataBuffer = calloc(1, size);
    
    // Now load all the data we want to save into this buffer
    uint8_t *ptr = keysDataBuffer;
    memcpy(ptr, nameKey, sizeof(nameKey));          // Load identifier
    ptr = ptr + sizeof(nameKey);
    memcpy(ptr, keys->keys_own.p_enc_key, sizeof(ble_gap_enc_key_t)); // load our LTK
    ptr = ptr + sizeof(ble_gap_enc_key_t);
    memcpy(ptr, keys->keys_own.p_id_key, sizeof(ble_gap_id_key_t));   // load our LTK id
    ptr = ptr + sizeof(ble_gap_id_key_t);
    memcpy(ptr, keys->keys_peer.p_enc_key, sizeof(ble_gap_enc_key_t)); // load peer LTK
    ptr = ptr + sizeof(ble_gap_enc_key_t);
    memcpy(ptr, keys->keys_peer.p_id_key, sizeof(ble_gap_id_key_t));   // load peer LTK id
    ptr = ptr + sizeof(ble_gap_id_key_t);
    memcpy(ptr, noOfCccds, sizeof(unsigned short));             // Load number of CCCDs
    ptr = ptr + sizeof(unsigned short);
    memcpy(ptr, cccdSet, sizeof(unsigned char) * (*noOfCccds)); // Load the cccdSet[]
    ptr = ptr + sizeof(unsigned char) * (*noOfCccds);
    memcpy(ptr, saveDataLength, sizeof(unsigned short));        // Load saveDataLength
    ptr = ptr + sizeof(unsigned short);
    if (*saveDataBuffer != NULL)
    {
        memcpy(ptr, *saveDataBuffer, *saveDataLength * sizeof(unsigned char));   // Load the saveDataBuffer
        ptr = ptr + *saveDataLength * sizeof(unsigned char);
    }

//    memset(buffer, 0 , 1024);
//    NRF_LOG_DEBUG("Full set of data being written %s", (uint32_t)byteToHex(keysDataBuffer, buffer, " ", size));
//    NRF_LOG_FLUSH();

    // Now we have to write the data in hunks into flash
    // Each page is 1024 bytes, and a write is in 4-byte hunks
    // So we will find the number of 1024 byte pages to write,
    // and then the number of 4-byte hunks left over.

    // Disable soft device so we don't have to deal with events which would make this write PAINFULLY complicated to implement
    //sd_softdevice_disable();
    nrf_sdh_disable_request();
    // Buffer to write to flash. Need to write it in four-byte hunks
    uint32_t *ptr32 = (uint32_t *)keysDataBuffer;
    while (true)
    {
        // Where to write
        addr = (uint32_t *)(pg_size * pg_num);
        // Erase page:
        while(true)
        {
            err_code = sd_flash_page_erase(pg_num);
            if (err_code == NRF_SUCCESS)
            {
                break;
            }
            if (err_code != NRF_ERROR_BUSY)
            {
                NRF_LOG_DEBUG("Erasing data returned error %u.", err_code);
                APP_ERROR_CHECK(err_code);      // Program dies
            }
        }

        i = (size >= pg_size) ? (pg_size >> 2) : // Do as many write that will fill the page, eg if pg size = 1024, then 256 writes
                                (size >> 2);     // All fits in one page, so divide size by four
        NRF_LOG_DEBUG("Number of 4-byte hunks to write %u.", i);
        while(true)
        {
            err_code = sd_flash_write(addr,     // Pointer to start of flash location to be written.
                                      ptr32,    // Pointer to buffer with data to be written.
                                      i);       // Number of 32-bit words to write. Maximum size is the number of words in one
                                                // flash page. See the device's Product Specification for details (pg size I assume).
            if (err_code == NRF_SUCCESS)
            {
                break;
            }
            if (err_code != NRF_ERROR_BUSY)
            {
                NRF_LOG_ERROR("Writing data returned Error: %s (0x%04X)", getErrorString(err_code), err_code);
                APP_ERROR_CHECK(err_code);      // Program dies
            }
        }
        size = size - pg_size;  // Subtract a page size from the total size
        if (size <= 0)          // if zero or less, all data has been written
        {
            NRF_LOG_DEBUG("Flash written");
            break;
        }
        pg_num++;   // TODO we are on last page, so have to hope everything fits in one page. Still have not figured out how to get
                    // to the start of the first free page available.
        ptr32 = (uint32_t *)(keysDataBuffer + pg_size);
    }
    free(keysDataBuffer);
}

void loadKeysFromFlash(ble_gap_sec_keyset_t* keys,
    unsigned char** saveDataBuffer, unsigned short int* saveDataLength,
    unsigned char* cccdSet, unsigned short *noOfCccds)
{
    char localNameKey[8];
//        char buffer[1024];

    uint32_t pg_size = NRF_FICR->CODEPAGESIZE;
    uint32_t pg_num  = NRF_FICR->CODESIZE - (NRF_FICR->CODESIZE * pg_size - 0xE0000) / pg_size;
    
    // Okay, what are we doing here. On the dongle, the bootloader occupies flash above 0xE0000. So I have to find how big flash is,
    // subtract 0xE0000 from it, find the number of pages there, and then subtract that number of pages from the total. THAT gives the
    // page where the boot record starts (0xE0000). Now I have to drop a few pages to where I can start. I drop two pages here which is
    // more than enough room for this app. I still don't know how to dynamically find the page at the top of MY app which gives the
    // area of free flash between my app and the bootloader at the top. The DK, by the way, has no bootloader so I don't need all this.

    NRF_LOG_INFO("Total pages is %u, pg_num is %u pg_size is %u", NRF_FICR->CODESIZE, pg_num, NRF_FICR->CODEPAGESIZE);
    pg_num = pg_num - 2;

    uint8_t *addr = (uint8_t *)(pg_size * pg_num);
//    uint8_t *start = addr;

    memcpy(localNameKey, addr, sizeof(localNameKey));
    if (memcmp(localNameKey, nameKey, sizeof(nameKey)) != 0)
    {
        NRF_LOG_DEBUG("Flash has different name key - It is invalid or empty.");
        return;     // nothing written to Flash yet
    }
    addr = addr + sizeof(nameKey);
    memcpy(keys->keys_own.p_enc_key, addr, sizeof(ble_gap_enc_key_t));  // memcpy (destination, source, number of bytes to copy)
    addr = addr + sizeof(ble_gap_enc_key_t);
    memcpy(keys->keys_own.p_id_key, addr, sizeof(ble_gap_id_key_t));
    addr = addr + sizeof(ble_gap_id_key_t);
    memcpy(keys->keys_peer.p_enc_key, addr, sizeof(ble_gap_enc_key_t));
    addr = addr + sizeof(ble_gap_enc_key_t);
    memcpy(keys->keys_peer.p_id_key, addr, sizeof(ble_gap_id_key_t));
    addr = addr + sizeof(ble_gap_id_key_t);
    memcpy(noOfCccds, addr, sizeof(unsigned short));
    addr = addr + sizeof(unsigned short);
    memcpy(cccdSet, addr, sizeof(unsigned char) * (*noOfCccds)); // Load the cccdSet[]
    addr = addr + sizeof(unsigned char) * (*noOfCccds);
    memcpy(saveDataLength, addr, sizeof(unsigned short));        // Load saveDataLength
    addr = addr + sizeof(unsigned short);
    if (*saveDataLength > 0)
    {
        *saveDataBuffer = calloc(1, *saveDataLength * sizeof(unsigned char));
        memcpy(*saveDataBuffer, addr, *saveDataLength * sizeof(unsigned char)); // Load the saveDataBuffer
    }
//    memset(buffer, 0 , 1024);
//    unsigned long count = (unsigned long)(addr -start) + *saveDataLength;
//    NRF_LOG_DEBUG("Full set of data being read %s", (uint32_t)byteToHex(start, buffer, " ", count));
//    NRF_LOG_FLUSH();
}

void allocateMemoryForSecurityKeys(ble_gap_sec_keyset_t* keys)
{
    keys->keys_own.p_enc_key = calloc(1, sizeof(ble_gap_enc_key_t));
    keys->keys_own.p_id_key = calloc(1, sizeof(ble_gap_id_key_t));
    keys->keys_own.p_sign_key = calloc(1, sizeof(ble_gap_sign_info_t));
    keys->keys_own.p_pk = calloc(1, sizeof(ble_gap_lesc_p256_pk_t));

    keys->keys_peer.p_enc_key = calloc(1, sizeof(ble_gap_enc_key_t));
    keys->keys_peer.p_id_key = calloc(1, sizeof(ble_gap_id_key_t));
    keys->keys_peer.p_sign_key = calloc(1, sizeof(ble_gap_sign_info_t));
    keys->keys_peer.p_pk = calloc(1, sizeof(ble_gap_lesc_p256_pk_t));
}

void freeMemoryForSecurityKeys(ble_gap_sec_keyset_t* keys)
{
    if (keys->keys_own.p_enc_key != NULL) free(keys->keys_own.p_enc_key);
    if (keys->keys_own.p_id_key != NULL) free(keys->keys_own.p_id_key);
    if (keys->keys_own.p_sign_key != NULL) free(keys->keys_own.p_sign_key);
    if (keys->keys_own.p_pk != NULL) free(keys->keys_own.p_pk);
    
    if (keys->keys_peer.p_enc_key != NULL) free(keys->keys_peer.p_enc_key);
    if (keys->keys_peer.p_id_key != NULL) free(keys->keys_peer.p_id_key);
    if (keys->keys_peer.p_sign_key != NULL) free(keys->keys_peer.p_sign_key);
    if (keys->keys_peer.p_pk != NULL) free(keys->keys_peer.p_pk);
}


